/*AM1 to x86 Assembly Translator C++ version
Copyright (C) 2009 Felix Schmitt

feel free to report about this software to s0760034@mail.zih.tu-dresden.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.*/

/* 
* translates to x86 for linux and win32 systems
* error checking is not complete yet
* mixed c/c++ code
*
* compile this file with:
* g++ -o am1toasmC[.exe] am1toasmC.cpp
*
* create asm code with:
* am1toasmC[.exe] infile file.asm [-i] [-elf/-win32]
*
* compile and link asm code with:
* nasm -f elf/win32 -o file.o/file.obj file.asm
* gcc -o outfile file.o/file.obj
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <vector>

using namespace std;

// some constants

#define CONSOLE_MSG "am1toasm translator, Copyright (C) 2009 Felix Schmitt\n"\
	"am1toasm comes with ABSOLUTELY NO WARRANTY;\nThis is free software, and you are welcome\n"\
	"to redistribute it under certain conditions.\n"\
	"usage: am1toasm (input) (output) [-i] [-elf/-win32]"

#define CMP_CODE " ltrue_%d\n\tpush dword 1\n\tjmp lfalse_%d\nltrue_%d:\n\tpush dword 0\nlfalse_%d:"

#define STATIC_DEF "%define main _main\n"\
	"%define scanf _scanf\n"\
	"%define printf _printf\n"\
	"%define putchar _putchar\n"

#define STATIC_HEADER "segment .text\n\t"\
	"global main, print_int, read_int, print_string, print_nl\n\t"\
	"extern printf, scanf, putchar"

#define STATIC_CODE "print_int:\nenter 0,0\npusha\npushf\n"\
	"push eax\npush dword format_int\ncall printf\npop ecx\npop ecx\npopf\npopa\nleave\nret\n"\
	"read_int:\nenter 4,0\npusha\npushf\nlea eax, [ebp-4]\npush eax\n"\
	"push dword format_int\ncall scanf\npop ecx\npop ecx\npopf\npopa\nmov eax, [ebp-4]\nleave\nret\n"\
	"print_string:\nenter 0,0\npusha\npushf\n"\
	"push eax\npush dword format_string\ncall printf\npop ecx\npop ecx\npopf\npopa\nleave\nret\n"\
	"print_nl:\nenter 0,0\npusha\npushf\npush dword 10\ncall putchar\npop	ecx\npopf\npopa\nleave\nret\n"\
	"segment .data\n\tformat_int db \"%i\",0\n\tformat_string db \"%s\",0\n"\
	"\tmsg_in db \"Enter number: \",0\n\tmsg_out db \"Result: \",0"

const int FUNC_GLOBAL = 1;
const int VAR_SIZE = 4;
const int BUFFER_SIZE = 256;
const int ARCH_LINUX = 0;
const int ARCH_WIN = 1;

// represents current translation context
struct Context
{
	int funcID;
	int localStackSize;
	int parStackSize;
	int globalVars;
	int architecture;
};

// writes code for comparison statements to output file
void writeCmpCode(int lineNumber, FILE *file)
{
	char lineNumStr[BUFFER_SIZE];
	sprintf(lineNumStr, "%d", lineNumber);
	char *buffer = new char[strlen(CMP_CODE) - 8 + 4 * strlen(lineNumStr) + 1];

	sprintf(buffer, CMP_CODE, lineNumber, lineNumber, lineNumber, lineNumber);
	fputs(buffer, file);

	free(buffer);
}

// includes static header in output file
void writeStaticHeader(FILE *file)
{
	fputs(STATIC_HEADER, file);
}

// appends static code to output file
void writeStaticCode(FILE *file)
{
	fputs(STATIC_CODE, file);
}

// traces source code for jmp/jmc codes and saves line numbers
// returns -1 on failure
int createJumpLabels(FILE *file, vector<int> *labels)
{
	if (file == NULL)
		return -1;

	char buffer[BUFFER_SIZE];
	int lineNumber = 0;

	while (fgets(buffer, BUFFER_SIZE, file))
	{
		lineNumber++;
		for (int i = 0; i < strlen(buffer) - 1; i++)
			buffer[i] = toupper(buffer[i]);

		if (strlen(buffer) > 5 && (strstr(buffer, "JMP") != NULL || strstr(buffer, "JMC") != NULL))
		{
			char *label;

			label = strtok(buffer," ;");
			if (label != NULL)
			{
				label = strtok(NULL, " ;");
				labels->push_back(atoi(label));
			}
			else
				return -1;
		}
	}

	return 0;
}

// identifies opcode by number, -1 on failure
int getOpcodeNum(char *opcode)
{
	const int codes_len = 27;
	const char* codes[codes_len] = {"PUSH", "ADD", "SUB", "MUL", "DIV", "MOD", "EQ", "GT",
		"LT", "GE", "LE", "NE", "CALL", "JMP", "JMC", "INIT", "LIT", "RET",
		"LOADI", "STOREI", "READI", "WRITEI", "LOAD", "LOADA", "STORE", "READ",
		"WRITE"};

	for (int i = 0; i < codes_len; i++)
		if (strcmp(opcode, codes[i]) == 0)
			return i;

	return -1;
}


// translates one line of am1 code to assembler
// returns 0 on success, -1 otherwise
int translateLine(int lineNumber, char *bufLine, FILE *file, Context *context, vector<int> jumpLabels)
{
	char *opcodeBuf;
	char *line;
	vector<char*>opcodes;
	int bufLineLen = strlen(bufLine);

	line = new char[bufLineLen];
	memset(line, 0, bufLineLen);

	if (bufLine[bufLineLen - 1] == ';')
		strncpy(line, bufLine, bufLineLen - 1);
	else
		if (bufLine[bufLineLen - 2] == ';') 
			strncpy(line, bufLine, bufLineLen - 2);
		else
		{
			printf("error: invalid opcode '%s' (missing ';' ?)\n", bufLine);
			return -1;
		}

	opcodeBuf = strtok(line, " .,()");
	while (opcodeBuf != NULL)
	{
		if (strcmp(opcodeBuf, "") != 0)
			opcodes.push_back(opcodeBuf);
		opcodeBuf = strtok(NULL, " .,()");
	}

	int result = 0;
	int opcodeNum = getOpcodeNum(opcodes[0]);
	if (opcodeNum == -1)
		result = -1;

	int opcode2;
	int opcode3;
	char buf[BUFFER_SIZE];

	if ((opcodeNum >= 12 && opcodes.size() < 2) ||
		(opcodeNum >= 22 && opcodes.size() < 3) ||
		(opcodeNum >= 12 && opcodeNum <= 21 && opcodes.size() > 2) ||
		(opcodes.size() >= 4))
	{
		printf("error: invalid opcode '%s'\n", line);
		result = -1;
	}

	if (result != -1)
		switch (opcodeNum)
	{
		case 0: //PUSH
			context->parStackSize += VAR_SIZE;
			fputs("\tnop", file);
			break;
		case 1: //ADD
			fputs("\tpop eax\n\tpop ebx\n\tadd ebx, eax\n\tpush ebx", file);
			break;
		case 2: //SUB
			fputs("\tpop eax\n\tpop ebx\n\tsub ebx, eax\n\tpush ebx", file);
			break;
		case 3: //MUL
			fputs("\tpop ebx\n\tpop eax\n\timul ebx\n\tpush eax", file);
			break;
		case 4: //DIV
			fputs("\tpop ebx\n\tpop eax\n\tmov edx, 0\n\tidiv ebx\n\tpush eax", file);
			break;
		case 5: //MOD
			fputs("\tpop ebx\n\tpop eax\n\tmov edx, 0\n\tidiv ebx\n\tpush edx", file);
			break;
		case 6: //EQ
			fputs("\tpop ebx\n\tpop eax\n\tcmp eax, ebx\n\tje", file);
			writeCmpCode(lineNumber, file);
			break;
		case 7: //GT
			fputs("\tpop ebx\n\tpop eax\n\tcmp eax, ebx\n\tjg", file);
			writeCmpCode(lineNumber, file);
			break;
		case 8: //LT
			fputs("\tpop ebx\n\tpop eax\n\tcmp eax, ebx\n\tjl", file);
			writeCmpCode(lineNumber, file);
			break;
		case 9: //GE
			fputs("\tpop ebx\n\tpop eax\n\tcmp eax, ebx\n\tjge", file);
			writeCmpCode(lineNumber, file);
			break;
		case 10: //LE
			fputs("\tpop ebx\n\tpop eax\n\tcmp eax, ebx\n\tjle", file);
			writeCmpCode(lineNumber, file);
			break;
		case 11: //NE
			fputs("\tpop ebx\n\tpop eax\n\tcmp eax, ebx\n\tjne", file);
			writeCmpCode(lineNumber, file);
			break;
		case 12: //CALL
			sprintf(buf, "\tcall func_%s\n\tadd esp, %d", opcodes[1], context->parStackSize);
			fputs(buf, file);
			break;
		case 13: //JMP
			if (context->funcID == FUNC_GLOBAL)
			{
				if (context->architecture == ARCH_WIN)
					fputs("\tmov esp, ebp\n\tpop ebp\n\tpush dword 0\n\tcall _ExitProcess@4", file);
				else
					fputs("\tmov esp, ebp\n\tpop ebp\n\tmov ebx, 0\n\tmov eax, 1\n\tint 80h", file);
			}
			else
			{
				fputs("\tjmp label_", file);
				fputs(opcodes[1], file);
			}
			break;
		case 14: //JMC
			fputs("\tpop ebx\n\tmov eax, 0\n\tcmp ebx, eax\n\tje label_", file);
			fputs(opcodes[1], file);
			break;
		case 15: //INIT
			context->funcID = lineNumber;
			opcode2 = atoi(opcodes[1]);

			if (context->funcID == FUNC_GLOBAL)
			{
				context->globalVars = opcode2;
				fputs("segment .data\n", file);
				for (int i = 0; i < context->globalVars; i++)
				{
					sprintf(buf, "\tglob%d dd 0\n", i + 1);
					fputs(buf, file);
				}

				writeStaticHeader(file);
				if (context->architecture == ARCH_WIN)
					fputs(", _ExitProcess@4", file);

				fputs("\nmain:\n\tpush ebp\n\tmov ebp, esp", file);
			}
			else
			{
				context->localStackSize = opcode2 * VAR_SIZE;
				sprintf(buf, "func_%d:\n\tpush ebp\n\tmov ebp, esp\n\tsub dword esp, %d",
					context->funcID, context->localStackSize);
				fputs(buf, file);
			}
			break;
		case 16: //LIT
			fputs("\tpush dword ", file);
			fputs(opcodes[1], file);
			break;
		case 17: //RET
			fputs("\tmov esp, ebp\n\tpop ebp\n\tret", file);
			break;
		case 18: //LOADI
			opcode2 = atoi(opcodes[1]);

			if (opcode2 < 0)
				// is parameter
				fputs("\tmov eax, [ebp+", file);
			else
				// is local variable
				fputs("\tmov eax, [ebp-", file);

			sprintf(buf, "%d", abs(opcode2) * VAR_SIZE);
			fputs(buf, file);
			fputs("]\n\tpush dword [eax]", file);
			break;
		case 19: //STOREI
			opcode2 = atoi(opcodes[1]);

			if (opcode2 < 0)
				// is parameter
				fputs("\tmov eax, [ebp+", file);
			else
				// is local variable
				fputs("\tmov eax, [ebp-", file);

			sprintf(buf, "%d", abs(opcode2) * VAR_SIZE);
			fputs(buf, file);
			fputs("]\n\tpop dword [eax]", file);
			break;
		case 20: //READI
			opcode2 = atoi(opcodes[1]);

			fputs("\tmov eax, msg_in\n\tcall print_string\n\tcall read_int\n\t", file);

			if (opcode2 < 0)
				// is parameter
				fputs("mov ebx, [ebp+", file);
			else
				// is local variable
				fputs("mov ebx, [ebp-", file);

			sprintf(buf, "%d", abs(opcode2) * VAR_SIZE);
			fputs(buf, file);
			fputs("]\n\tmov [ebx], eax", file);
			break;
		case 21: //WRITEI
			opcode2 = atoi(opcodes[1]);

			fputs("\tmov eax, msg_out\n\tcall print_string\n\t", file);

			if (opcode2 < 0)
				// is parameter
				fputs("mov ebx, [ebp+", file);
			else
				// is local variable
				fputs("mov ebx, [ebp-", file);

			sprintf(buf, "%d", abs(opcode2) * VAR_SIZE);
			fputs(buf, file);
			fputs("]\n\tmov eax, [ebx]\n\tcall print_int\n\tcall print_nl", file);
			break;
		case 22: //LOAD
			if (strcmp(opcodes[1], "GLOBAL") == 0)
			{
				fputs("\tpush dword [glob", file);
				fputs(opcodes[2], file);
				fputs("]", file);
			}
			else
			{
				opcode3 = atoi(opcodes[2]);
				if (opcode3 < 0)
					// is parameter
					fputs("\tpush dword [ebp+", file);
				else
					// is local variable
					fputs("\tpush dword [ebp-", file);

				sprintf(buf, "%d", abs(opcode3) * VAR_SIZE);
				fputs(buf, file);
				fputs("]", file);
			}
			break;
		case 23: //LOADA
			if (strcmp(opcodes[1], "GLOBAL") == 0)
			{
				fputs("\tpush dword glob", file);
				fputs(opcodes[2], file);
			}
			else
			{
				opcode3 = atoi(opcodes[2]);
				fputs("\tmov eax, ebp\n\tsub dword eax, ", file);

				sprintf(buf, "%d", abs(opcode3) * VAR_SIZE);
				fputs(buf, file);
				fputs("\n\tpush eax", file);
			}
			break;
		case 24: //STORE
			if (strcmp(opcodes[1], "GLOBAL") == 0)
			{
				fputs("\tpop dword [glob", file);
				fputs(opcodes[2], file);
				fputs("]", file);
			}
			else
			{ 
				opcode3 = atoi(opcodes[2]);
				if (opcode3 < 0)
					// is parameter
					fputs("\tpop dword [ebp+", file);
				else
					// is local variable
					fputs("\tpop dword [ebp-", file);

				sprintf(buf, "%d", abs(opcode3) * VAR_SIZE);
				fputs(buf, file);
				fputs("]", file);
			}
			break;
		case 25: //READ
			if (strcmp(opcodes[1], "GLOBAL") == 0)
			{
				fputs("\tmov eax, msg_in\n\tcall print_string\n\tcall read_int\n\tmov [glob", file);
				fputs(opcodes[2], file);
				fputs("], eax", file);
			}
			else
			{
				opcode3 = atoi(opcodes[2]);
				fputs("\tmov eax, msg_in\n\tcall print_string\n\tcall read_int\n", file);

				if (opcode3 < 0)
					// is parameter
					fputs("\tmov dword [ebp+", file);
				else
					// is local variable
					fputs("\tmov dword [ebp-", file);

				sprintf(buf, "%d", abs(opcode3) * VAR_SIZE);
				fputs(buf, file);
				fputs("], eax", file);
			}
			break;
		case 26: //WRITE
			if (strcmp(opcodes[1], "GLOBAL") == 0)
			{
				fputs("\tmov eax, msg_out\n\tcall print_string\n\tmov dword eax, [glob", file);
				fputs(opcodes[2], file);
				fputs("]\n\tcall print_int\n\tcall print_nl", file);
			}
			else
			{
				opcode3 = atoi(opcodes[2]);
				fputs("\tmov eax, msg_out\n\tcall print_string\n\t", file);

				if (opcode3 < 0)
					// is parameter
					fputs("mov dword eax, [ebp+", file);
				else
					// is local variable
					fputs("mov dword eax, [ebp-", file);

				sprintf(buf, "%d", abs(opcode3) * VAR_SIZE);
				fputs(buf, file);
				fputs("]\n\tcall print_int\n\tcall print_nl", file);
			}
			break;
		default:
			// unknown opcode
			printf("error: invalid opcode '%s'\n", line);
			result = -1;
			break;
	}

	fputs("\n", file);

	free(line);

	return result;
}

// translates am1 code to assembler,
// returns -1 on failure
int assemble(FILE *inFile, FILE *outFile, bool incSrc, int arch)
{
	if (inFile == NULL || outFile == NULL)
	{
		printf("error: file error\n");
		return -1;
	}

	char buffer[BUFFER_SIZE];
	Context context;
	context.funcID = FUNC_GLOBAL;
	context.globalVars = 0;
	context.localStackSize = 0;
	context.parStackSize = 0;
	context.architecture = arch;

	int lineNumber = 0;

	vector<int> jumpLabels;

	if (createJumpLabels(inFile, &jumpLabels) != 0)
		return -1;

	rewind(inFile);

	if (arch == ARCH_WIN)
		fputs(STATIC_DEF, outFile);

	while (fgets(buffer, BUFFER_SIZE, inFile))
	{
		if (strcmp(buffer, "") != 0 && strcmp(buffer, "\n") != 0)
		{
			lineNumber++;
			if (incSrc)
			{
				fputs("; ", outFile);
				fputs(buffer, outFile);
				fputs("\n", outFile);
			}

			for (int i = 0; i < jumpLabels.size(); i++)
			{
				if (jumpLabels[i] == lineNumber)
				{
					char labelBuf[BUFFER_SIZE];
					fputs("label_", outFile);
					sprintf(labelBuf, "%d", jumpLabels[i]);
					fputs(labelBuf, outFile);
					fputs(":\n", outFile);
					break;
				}
			}

			for (int i = 0; i < strlen(buffer) - 1; i++)
				buffer[i] = toupper(buffer[i]);

			if (translateLine(lineNumber, buffer, outFile, &context, jumpLabels) != 0)
			{
				printf("error: unable to translate line %d: %s", lineNumber, buffer);
				return -1;
			}
		}
	}

	writeStaticCode(outFile);
	return 0;
}

int main(int argc, char* argv[])
{
	if (argc < 3)
	{
		printf("%s\n", CONSOLE_MSG);
		printf("too few arguments\n");
		return 0;
	}

	FILE *inFile = fopen(argv[1], "r");

	if (inFile != NULL)
	{
		FILE *outFile = fopen(argv[2], "w");

		if (outFile != NULL)
		{
			// include src code in output file
			bool incSrc = false;
			int arch = ARCH_LINUX;

			// test arguments
			if (argc > 3)
			{
				for (int i = 3; i < argc; i++)
				{
					if (strcmp(argv[i], "-i") == 0)
						incSrc = true;

					if (strcmp(argv[i], "-win32") == 0)
						arch = ARCH_WIN;

					if (strcmp(argv[i], "-elf") == 0)
						arch = ARCH_LINUX;
				}
			}

			if (arch == ARCH_WIN)
				printf("target is win32\n");
			else
				printf("target is linux\n");

			if (assemble(inFile, outFile, incSrc, arch) != 0)
				printf("\nfailed\n");
			else
				printf("\ndone\n");

			fclose(outFile);
		}
		else
			printf("could not open output file\n");

		fclose(inFile);
	}
	else
		printf("could not read input file\n");

	return 0;
}

